package hexagon;

import java.util.function.Consumer;

import org.lwjgl.input.Keyboard;
import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector3f;

import hexagon.graphics.Shader;
import hexagon.util.CylVec3;
import hexagon.util.MathUtil;
import hexagon.util.Vec3f;
import hexagon.world.World;

public class Camera {
	
	public float fov, aspect, near, far, camX;
	public Vec3f position;
	public Vector3f blockCoords, placeInBlock;
	public Vector3f rotation;
	public Matrix4f projectionMatrix, viewMatrix;
	
	public boolean fly = false, F_down;
	public CylVec3 playerVelocity;
	private World world;
	
	private Consumer<Shader> setViewMatrix = s -> {
		s.setUniformMat4("viewMatrix", viewMatrix);
		s.setUniform1f("camX", camX);
	};
	private Consumer<Shader> setProjMatrix = s -> s.setUniformMat4("projectionMatrix", projectionMatrix);
	
	public Camera(float fov, float aspect, float near, float far, World world) {
		this.fov = fov;
		this.aspect = aspect;
		this.near = near;
		this.far = far;
		this.world = world;
		position = new Vec3f();
		blockCoords = new Vector3f();
		placeInBlock = new Vector3f();
		rotation = new Vector3f();
		playerVelocity = new CylVec3();
		projectionMatrix = MathUtil.projectionMatrix(fov, aspect, near, far, null);
		viewMatrix = MathUtil.viewMatrix(this, null);
	}
	
	public void update() {
		CylVec3 cylPos = position.getCyl();
		float speed = Keyboard.isKeyDown(Keyboard.KEY_LCONTROL) ? 0.01f : Keyboard.isKeyDown(Keyboard.KEY_RCONTROL) ? 10f : 0.5f, rSpeed = 0.05f;
		speed *= cylPos.r / World.radius;
		if (Keyboard.isKeyDown(Keyboard.KEY_W)) {
			playerVelocity.rv += (float) (Math.cos(rotation.y) * speed);
			playerVelocity.x += (float) (Math.sin(rotation.y) * speed);
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_S)) {
			playerVelocity.rv -= (float) (Math.cos(rotation.y) * speed);
			playerVelocity.x -= (float) (Math.sin(rotation.y) * speed);
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_D)) {
			playerVelocity.x += (float) (Math.cos(rotation.y) * speed);
			playerVelocity.rv -= (float) (Math.sin(rotation.y) * speed);
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_A)) {
			playerVelocity.x -= (float) (Math.cos(rotation.y) * speed);
			playerVelocity.rv += (float) (Math.sin(rotation.y) * speed);
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_SPACE)) {
			if (fly) cylPos.rv /= cylPos.r / (cylPos.r += speed);
			else if (playerVelocity.r == 0) playerVelocity.r += speed;
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_LSHIFT)) {
			if (fly) cylPos.rv /= cylPos.r / (cylPos.r -= speed);
		}
		
		float velocityFactor = (float) Math.hypot(playerVelocity.x, playerVelocity.rv) / speed;
		if (velocityFactor > 1) {
			playerVelocity.x /= velocityFactor;
			playerVelocity.rv /= velocityFactor;
		}
		if (!fly) {
			playerVelocity.r -= 9.82f / 3600;
			
			if (placeInBlock.y > 0.1f && world.getBlock((int) blockCoords.x, (int) blockCoords.y + 3, (int) blockCoords.z) != null) {
				if (playerVelocity.r < 0) playerVelocity.r = 0;
			}
		}
		
		cylPos.rv /= cylPos.r / (cylPos.r += playerVelocity.r);
		cylPos.rv += playerVelocity.rv;
		cylPos.x += playerVelocity.x;
		
		playerVelocity.x *= 0.75f;
		playerVelocity.rv *= 0.75f;
		
		if (Keyboard.isKeyDown(Keyboard.KEY_UP)) {
			rotation.x -= rSpeed;
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_DOWN)) {
			rotation.x += rSpeed;
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_LEFT)) {
			rotation.y -= rSpeed;
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_RIGHT)) {
			rotation.y += rSpeed;
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_PRIOR)) {
			rotation.z -= rSpeed;
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_NEXT)) {
			rotation.z += rSpeed;
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_DELETE)) {
			if (Keyboard.isKeyDown(Keyboard.KEY_P)) {
				cylPos.x = 0;
				cylPos.rv = 0;
				cylPos.r = (float) World.radius + 2;
			}
			if (Keyboard.isKeyDown(Keyboard.KEY_R)) {
				rotation.x = rotation.y = rotation.z = 0;
			}
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_F)) {
			if (!F_down) {
				F_down = true;
				fly = !fly;
			}
		} else F_down = false;
		
		camX = position.getCyl().x;
		position.updateEucl();
		MathUtil.toBlockCoords(position.getCyl(), blockCoords, placeInBlock);
		MathUtil.viewMatrix(this, viewMatrix);
		Shader.forEachShader(setViewMatrix);
	}
	
	public void loadProjMatrToShader(Shader shader) {
		shader.enable();
		setProjMatrix.accept(shader);
	}
	
}
