package hexagon;

import java.io.File;

import org.lwjgl.input.Keyboard;
import org.lwjgl.input.Mouse;
import org.lwjgl.opengl.ContextAttribs;
import org.lwjgl.opengl.Display;
import org.lwjgl.opengl.DisplayMode;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.PixelFormat;
import org.newdawn.slick.util.Log;

import hexagon.graphics.Shader;
import hexagon.model.HexagonModel;
import hexagon.model.Loader;
import hexagon.util.MousePicker;
import hexagon.world.World;

public class Main {
	
	/**
	 *	Jag kanske skulle kunna göra tutorials på att programmera detta spel som ett gymnasiearbete! 
	 */
	
	public static final int width = 854, height = 480;
	public static MousePicker mousePicker;
	public static Main main;
	
	private World world;
	public Camera camera;
	
	private boolean F3_down = false;
	
	public Main() {
		main = this;
		initGL();
		initGame();
	}
	
	private void initGame() {
		HexagonModel.init();
		Shader.init();
		
		world = new World();
		
		camera = new Camera(70f, 16f / 9f, 0.02f, 1000f, world);
		camera.position.getCyl().r = (float) World.radius + 4;
		camera.position.updateEucl();
		camera.loadProjMatrToShader(Shader.block);
		Shader.block.setUniform1f("radius", (float) World.radius);
		mousePicker = new MousePicker(camera);
	}
	
	private void tick() {
		if (Keyboard.isKeyDown(Keyboard.KEY_F3)) {
			if (!F3_down) {
				F3_down = true;
				Shader.init();
				camera.loadProjMatrToShader(Shader.block);
				Shader.block.setUniform1f("radius", (float) World.radius);
				Loader.theLoader.clearTextures();
				HexagonModel.init();
				System.out.println("Reloaded textures and shaders.");
			}
		} else F3_down = false;
		
		camera.update();
		mousePicker.update();
		
		world.tick(camera);
	}
	
	private void loop() {
		long prevTime = System.nanoTime(), currentTime, realPrevTime = prevTime, realDeltaTime = 0, totalPerSecond = 0;
		int delta, ticks = 0, frames = 0, fps = 0;
		while (!Display.isCloseRequested()) {
			delta = (int) (((currentTime = System.nanoTime()) - prevTime) * 1e-9 * 60);
			realPrevTime = currentTime;
			for (int i = 0; i < delta; i++) {
				tick();
				ticks++;
				if (ticks % 60 == 0) {
//					if (ticks > 300) System.out.println("ms/update: " + totalPerSecond * 1e-6f / fps);
					fps = frames;
					frames = 0;
//					totalPerSecond = 0;
				}
			}
			prevTime += delta * 1e9 / 60;
			
			GL11.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);
			world.render(camera);
			
			int error = GL11.glGetError();
			if (error != GL11.GL_NO_ERROR) System.out.println("OpenGL error: " + error);
			
			String debugInfo = "    C: (" + camera.position.getCyl().x + ", " + camera.position.getCyl().r + ", " + camera.position.getCyl().rv + ")    E: (" + camera.position.getEucl().x + ", " + camera.position.getEucl().y + ", " + camera.position.getEucl().z + ")    R: (" + (float) (camera.rotation.x * 180 / Math.PI) + ", " + (float) (camera.rotation.y * 180 / Math.PI) + ", " + (float) (camera.rotation.z * 180 / Math.PI) + ")";
			Display.update();
//			totalPerSecond += 
			realDeltaTime = System.nanoTime() - realPrevTime;
			frames++;
			Display.setTitle(fps + " fps   ms: " + (int) (realDeltaTime * 1e-6) + debugInfo);
//			Display.sync(60);
		}
	}
	
	private void initGL() {
		try {
			Display.setDisplayMode(new DisplayMode(width, height));
			Display.create(new PixelFormat(), new ContextAttribs(3, 3));
			Keyboard.create();
			Mouse.create();
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		GL11.glClearColor(0, 0, 0, 1);
		
		GL11.glEnable(GL11.GL_DEPTH_TEST);
		GL11.glDepthFunc(GL11.GL_LEQUAL);
	}
	
	public static void destroy() {
		Loader.theLoader.cleanUp();
		Shader.cleanUp();
		Keyboard.destroy();
		Mouse.destroy();
		Display.destroy();
	}
	
	public static void main(String[] args) {
		File file = new File("lib/natives");
		if (!file.exists()) file = new File("lib/natives-" + (System.getProperty("os.name").toLowerCase().contains("win") ? "windows" : "linux"));
		System.setProperty("org.lwjgl.librarypath", file.getAbsolutePath());
		Log.setVerbose(false);
		
		new Main().loop();
		destroy();
	}
	
}
