package hexagon.world;

import org.lwjgl.util.vector.Vector3f;

import hexagon.graphics.BlockRenderer;

public class Block {
	
	public static final byte[][] neighbourOffsets = {
			{0, -1, 0},
			{0, 1, 0},
			{1, 0, 0},
			{0, 0, 1},
			{-1, 0, 1},
			{-1, 0, 0},
			{0, 0, -1},
			{1, 0, -1}
	};
	
	public final int x, y, z;
	protected World world;
	public Vector3f colorTint;
	private byte sidesToRender;
	protected boolean willUpdate = false;
	
	public Block(int x, int y, int z, World world) {
		this.x = x;
		this.y = y;
		this.z = z;
		this.world = world;
		colorTint = new Vector3f(1f, 1f - (float) Math.random() * 0.5f, 1f);
		sidesToRender = ~0;
	}
	
	public void submit() {
		if (sidesToRender != 0) BlockRenderer.block.submit(this);
	}
	
	public boolean setRenderSide(byte side, boolean render) {
		int prev = sidesToRender;
		if (render) sidesToRender |= 1 << side; // set to 1
		else sidesToRender &= ~(1 << side); // set to 0
		
		return sidesToRender != prev;
	}
	
	public int setSidesToRender(byte sides) {
		int diff = sides ^ sidesToRender;
		sidesToRender = sides;
		
		return diff;
	}
	
	public byte sidesToRender() {
		return sidesToRender;
	}
	
	public void onRemove() {
		sidesToRender = (byte) ~sidesToRender;
		onUpdate(false);
	}
	
	public void onUpdate(boolean silent) {
		byte sides = 0;
		Block b;
		for (int i = 0; i < 8; i++) {
			b = world.getBlock(x + neighbourOffsets[i][0], y + neighbourOffsets[i][1], z + neighbourOffsets[i][2]);
			if (b == null) sides |= 1 << i;
			else if ((sidesToRender & 1 << i) != 0) world.addBlockToUpdateList(b);
		}
		sidesToRender = sides;
	}
	
}
