#version 330 core
#define PI 3.14159265
#define PId3 PI / 3

layout (points) in;
layout (triangle_strip, max_vertices = 36) out;

in VertOut {
	vec3 colorTint;
	int sides;
} vertOut[];

out FragIn {
	vec2 texCoords;
	vec3 colorTint;
} fragIn;

uniform mat4 projectionMatrix;
uniform mat4 viewMatrix;
uniform float angleHalfHexagon;
uniform float radius;
uniform float camX;
uniform float camY;

void main() {
	fragIn.colorTint = vertOut[0].colorTint;
	fragIn.texCoords = vec2(0, 0);
	mat4 matrix = projectionMatrix * viewMatrix;
	vec4 pos = gl_in[0].gl_Position;
	int sides = vertOut[0].sides;
	
	vec4 pos_map[12];
	vec2 tex_map[6];
	int index_map[6];
	
	float y60 = 0.8660254, x60 = 0.5;
	float xMult = radius / (y60 / angleHalfHexagon);
	float scale = 1 - 1 / radius;
	float mult2 = pow(scale, (camY - pos.y) * xMult), mult1 = mult2 / pow(scale, xMult);
	float xpos = (x60 + 1) * pos.x;
	
	int index, i, idx = 0, side;
	
	for (i = 1; i > -2; i--) { // 1, 0, -1
		float v1 = (pos.x + 2 * pos.z + i) * angleHalfHexagon;
		float v2 = PId3 * i;
		float y = sin(v1) * radius, z = cos(v1) * radius, cosv2 = cos(v2), sinv2 = y60 * i;
		vec3 tempVec;
		
		index_map[idx] = index = (6 + i) % 6;
		tex_map[index] = vec2(cosv2, sinv2);
		tempVec = vec3((xpos + cosv2) * xMult - camX, y, z) * mult1;
		tempVec.x += camX;
		pos_map[index] = matrix * vec4(tempVec, 1);
		tempVec = vec3((xpos + cosv2) * xMult - camX, y, z) * mult2;
		tempVec.x += camX;
		pos_map[index + 6] = matrix * vec4(tempVec, 1);
		
		index_map[idx + 1] = index = (3 - i) % 6;
		tex_map[index] = vec2(-cosv2, sinv2);
		tempVec = vec3((xpos - cosv2) * xMult - camX, y, z) * mult1;
		tempVec.x += camX;
		pos_map[index] = matrix * vec4(tempVec, 1);
		tempVec = vec3((xpos - cosv2) * xMult - camX, y, z) * mult2;
		tempVec.x += camX;
		pos_map[index + 6] = matrix * vec4(tempVec, 1);
		idx += 2;
	}
	
	side = sides & 1;
	for (i = 0; i < 6; i++) {
		index = index_map[i];
		gl_Position = side * pos_map[index];
		fragIn.texCoords = vec2((1 + tex_map[index].x) / 2, (1 - tex_map[index].y) / 2);
		EmitVertex();
	}
	EndPrimitive();
	side = (sides & 1 << 1) >> 1;
	for (i = 0; i < 6; i++) {
		index = index_map[i];
		gl_Position = side * pos_map[index_map[5 - i] + 6];
		fragIn.texCoords = vec2((1 - tex_map[index].x) / 2, (1 - tex_map[index].y) / 2);
		EmitVertex();
	}
	EndPrimitive();
	for (int j = 2; j < 8; j++) {
		side = (sides & 1 << j) >> j;
		for (i = 0; i < 4; i++) {
			index = (j - 1 - i % 2) % 6 + 6 * (i / 2);
			gl_Position = side * pos_map[index];
			fragIn.texCoords = vec2(1 - i % 2, i / 2);
			EmitVertex();
		}
		EndPrimitive();
	}
}