package hexagon.world;

import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.TreeMap;

public class BlockColumn {
	
	protected TreeMap<Integer, Block> blocks;
	protected Chunk chunk;
	
	public BlockColumn(Chunk chunk) {
		this.chunk = chunk;
		blocks = new TreeMap<>();
	}
	
	public Block getBlock(int y) {
		return blocks.get(y);
	}
	
	public void setBlock(Block block, int y) {
		Block prev = blocks.put(y, block);
		if (prev != null) prev.onRemove(false);
		chunk.world.addBlockToUpdateList(block);
	}
	
	public void removeBlock(int y) {
		Block prev = blocks.remove(y);
		prev.onRemove(false);
	}
	
	public void onRemove(boolean noUpdate) {
		for (Block b : blocks.values())
			b.onRemove(noUpdate);
		blocks.clear();
		chunk = null;
	}
	
	protected void saveColumn(ByteBuffer to, FileChannel fc) {
		for (Block block : blocks.values()) {
			block.saveBlock(to, fc);
		}
	}
	
}
