package hexagon.world;

import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;

import org.lwjgl.util.vector.Vector3f;

import hexagon.util.MathUtil;

public class Block {
	
	public static final byte[][] neighbourOffsets = {
			{ 0, 1, 0},
			{ 0,-1, 0},
			{ 1, 0, 0},
			{ 0, 0, 1},
			{-1, 0, 1},
			{-1, 0, 0},
			{ 0, 0,-1},
			{ 1, 0,-1}
	};
	
	public final long coords;
	public final int x, y, z;
	protected World world;
	public Vector3f colorTint;
	protected byte sidesToRender;
	protected boolean willUpdate = false;
	
	public Block(int x, int y, int z, World world) {
		this.x = x;
		this.y = y;
		this.z = z;
		coords = MathUtil.encodeCoords(x, y, z);
		this.world = world;
		colorTint = new Vector3f(1f, 1f - (float) Math.random() * 0.5f, 1f);
		sidesToRender = 0;
	}
	
	public byte sidesToRender() {
		return sidesToRender;
	}
	
	public void onRemove(boolean noUpdate) {
		colorTint = null;
		if (!noUpdate) {
			Block b;
			for (int i = 0; i < 8; i++) {
				b = world.getBlock(x + neighbourOffsets[i][0], y + neighbourOffsets[i][1], z + neighbourOffsets[i][2]);
				if (b != null) world.addBlockToUpdateList(b);
			}
		}
		world = null;
	}
	
	public void onUpdate() {
		byte sides = 0;
		Block b;
		for (int i = 0; i < 8; i++) {
			b = world.getBlock(x + neighbourOffsets[i][0], y + neighbourOffsets[i][1], z + neighbourOffsets[i][2]);
			if (b == null) sides |= 1 << i;
			else if ((b.sidesToRender & 1 << (i < 2 ? 1 - i : (i - 2 + 3) % 6 + 2)) != 0) world.addBlockToUpdateList(b);
		}
		sidesToRender = sides;
	}
	
	protected void saveBlock(ByteBuffer to, FileChannel fc) {
		if (to.remaining() < 24) world.saveBufferFull.accept(fc);
		to.putInt(x);
		to.putInt(y);
		to.putInt(z);
		to.putFloat(colorTint.x);
		to.putFloat(colorTint.y);
		to.putFloat(colorTint.z);
	}
	
}
