package hexagon.world;

import java.util.TreeMap;
import java.util.function.Consumer;

import hexagon.util.MathUtil;

public class Chunk {
	/** Number of blocks in a chunk in one dimension (eg. 4 means 4x4 blocks) */
	public static final int chunkSize = 16;
	/** Number of chunks in a ring */
	public static final int ringSize = 256;
	public static final int totalSize = chunkSize * ringSize;
	
	protected World world;
	protected final TreeMap<Long, Block> blocks = new TreeMap<>();
	public final int x, z;
	
	public Chunk(int x, int z, Consumer<Chunk> generator, World world) {
		this.x = x;
		this.z = z;
		this.world = world;
		
		if (!loadFromFile())
			generator.accept(this);
	}
	
	public Block getBlock(int x, int y, int z) {
		return blocks.get(MathUtil.encodeCoords(x, y, z));
	}
	
	public Block setBlock(int x, int y, int z, Block block) {
		world.addBlockToUpdateList(block);
		world.blocks.put(MathUtil.encodeCoords(x, y, z), block);
		return blocks.put(MathUtil.encodeCoords(x, y, z), block);
	}
	
	public Block removeBlock(int x, int y, int z) {
		Block b = blocks.remove(MathUtil.encodeCoords(x, y, z));
		world.blocks.remove(MathUtil.encodeCoords(x, y, z));
		
		if (b.sidesToRender != 0) {
			for (int i = 0; i < 8; i++) {
				if ((b.sidesToRender & 1 << i) != 0)
					world.removeFromRenderList(i, b.coords);
			}
		}
		
		b.onRemove(false);
		
		return b;
	}
	
	public void onRemove(boolean noUpdate) {
		saveChunk();
		
		for (Block b : blocks.values()) {
			world.blocks.remove(MathUtil.encodeCoords(b.x, b.y, b.z));
			
			if (b.sidesToRender != 0) {
				for (int i = 0; i < 8; i++) {
					if ((b.sidesToRender & 1 << i) != 0) world.removeFromRenderList(i, b.coords);
				}
			}
			
			b.onRemove(false);
		}
		blocks.clear();
		world = null;
	}
	
	protected void saveChunk() {
/*		if (world.saveFolder != null) {
			world.bytes = 0;
			world.time = 0;
			long start = 0;
			try {
				File file = new File(world.saveFolder.getAbsolutePath() + "/chunks/");
				file.mkdirs();
				file = new File(file, x + "." + z + ".dat");
				if (file.exists()) file.delete();
				file.createNewFile();
				RandomAccessFile rFile = new RandomAccessFile(file, "rw");
				FileChannel fc = rFile.getChannel();
				
				/*for (BlockColumn col : blockCols) {
					col.saveColumn(world.saveBuffer, fc);
				}*/
/*		long t2 = 0;
		for (int j = 0; j < 1000; j++) {
			for (int i = 0; i < 256; i++) {
				world.saveBuffer.putInt(i);
			}
			t2 = System.nanoTime();
			world.saveBufferFull.accept(fc);
			start += System.nanoTime() - t2;
		}
		start -= world.time;
		
		long t1 = System.nanoTime();
		rFile.close();
		System.out.println(System.nanoTime() - t1);
	} catch (IOException e) {
		e.printStackTrace();
	}
	
	System.out.println(world.bytes + " bytes written in " + start + " ns (" + world.time + " ns). That's " + (int) (world.bytes / 8e3 / (start / 1e9) * 10) / 10f + " kB/s (" + (int) (world.bytes / 8e3 / (world.time / 1e9) * 10) / 10f + " kB/s)");
	
	}*/
	}
	
	/** Tries to load from file. @return Whether or not the file existed and the loading could be performed. */
	private boolean loadFromFile() {
/*		if (world.saveFolder != null) {
			world.bytes = 0;
			long start = System.nanoTime();
			try {
				File file = new File(world.saveFolder.getAbsolutePath() + "/chunks/" + x + "." + z + ".dat");
				if (!file.exists()) return false;
				file.createNewFile();
				RandomAccessFile rFile = new RandomAccessFile(file, "rw");
				FileChannel fc = rFile.getChannel();
				
					for (BlockColumn col : blockCols) {
						col.saveColumn(world.saveBuffer, fc);
					}
					world.saveBufferFull.accept(fc);
				
				rFile.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
			start = System.nanoTime() - start;
			System.out.println(world.bytes + " bytes read in " + start + " ns.");
			
			return true;
		}*/
		return false;
	}
	
}
