package hexagon.world;

import java.io.File;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.channels.FileChannel;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.TreeMap;
import java.util.function.BiFunction;
import java.util.function.Consumer;

import org.lwjgl.BufferUtils;
import org.lwjgl.input.Mouse;
import org.lwjgl.util.vector.Vector2f;

import hexagon.Camera;
import hexagon.Main;
import hexagon.graphics.BlockRenderer;
import hexagon.graphics.Shader;
import hexagon.model.HexagonModel;
import hexagon.util.MathUtil;

public class World { // TODO: Move stuff from Chunk and Block to World, and make BlockTypes
	
	public static final float viewDistance = 5f * (HexagonModel.y60 * 2); // Change first number only!
	public static final float reachDistance = 8f;
	public static final double angleHalfHexagon = Math.PI / Chunk.totalSize;
	public static final double radius = 100;//HexagonModel.y60 / angleHalfHexagon;
	public static final double xMult = radius / (HexagonModel.y60 / angleHalfHexagon);
	
	private TreeMap<Long, Chunk> chunks = new TreeMap<>();
	protected TreeMap<Long, Block> blocks = new TreeMap<>();
	private LinkedList<Block> blocksToUpdate = new LinkedList<>();
	private LinkedList<Block> blocksUpdating = new LinkedList<>();
	private Block hoverBlock;
	private int blockAlteringCountdown = 0;
	
	// Rendering blocks
	protected int[] blockRenderStrides = new int[8];
	protected ByteBuffer[] blockRenderBuffers = new ByteBuffer[8];
	protected ArrayList<TreeMap<Long, Integer>> blockToIndexMaps = new ArrayList<>(8);
	protected ArrayList<ArrayList<Long>> indexToBlockLists = new ArrayList<>(8);
	
	// Saving to file
	protected File saveFolder;
	protected ByteBuffer saveBuffer = BufferUtils.createByteBuffer(1024);
	protected long bytes = 0, time = 0;
	protected Consumer<FileChannel> saveBufferFull = (fc) -> {
		try {
			bytes += saveBuffer.position();
			saveBuffer.flip();
			long start = System.nanoTime();
			fc.write(saveBuffer);
			time += System.nanoTime() - start;
			saveBuffer.clear();
			
		} catch (IOException e) {
			e.printStackTrace();
		}
	};
	
	private BiFunction<Integer, Integer, Consumer<Chunk>> tempBlockGenerator = (x, z) -> c -> {
		for (int i = 0; i < Chunk.chunkSize; i++) {
			for (int j = 0; j < Chunk.chunkSize; j++) {
				if (Math.random() < 0.8) c.setBlock(x * Chunk.chunkSize + i, 0, z * Chunk.chunkSize + j, new Block(x * Chunk.chunkSize + i, 0, z * Chunk.chunkSize + j, this));
				if (Math.random() < 0.9) c.setBlock(x * Chunk.chunkSize + i, -1, z * Chunk.chunkSize + j, new Block(x * Chunk.chunkSize + i, -1, z * Chunk.chunkSize + j, this));
				c.setBlock(x * Chunk.chunkSize + i, -2, z * Chunk.chunkSize + j, new Block(x * Chunk.chunkSize + i, -2, z * Chunk.chunkSize + j, this));
				for (int k = 3; k < 3; k++) {
					c.setBlock(x * Chunk.chunkSize + i, -k, z * Chunk.chunkSize + j, new Block(x * Chunk.chunkSize + i, -k, z * Chunk.chunkSize + j, this));
				}
			}
		}
	};
	
	public World() {
		for (int i = 0; i < 8; i++)
			blockRenderStrides[i] = i < 2 ? BlockRenderer.block.instanceSize : BlockRenderer.side.instanceSize;
		for (int i = 0; i < 8; i++)
			blockRenderBuffers[i] = ByteBuffer.allocate(10000 * blockRenderStrides[i]).order(ByteOrder.nativeOrder());
		for (int i = 0; i < 8; i++)
			blockToIndexMaps.add(new TreeMap<>());
		for (int i = 0; i < 8; i++)
			indexToBlockLists.add(new ArrayList<>());
	}
	
	public void tick(Camera camera) {
		updateHoverBlock();
		updateBlocks();
	}
	
	private void updateBlocks() {
		int numBlocks = blocksToUpdate.size();
		blocksUpdating.addAll(blocksToUpdate);
		
		for (Iterator<Block> it = blocksUpdating.iterator(); it.hasNext();) {
			Block b = it.next();
			if (b.world != null) updateBlock(b.x, b.y, b.z);
		}
		
		for (int i = 0; i < numBlocks; i++) {
			Block b = blocksToUpdate.removeFirst();
			b.willUpdate = false;
		}
		
		blocksUpdating.clear();
	}
	
	public void updateLoadedChunks(Camera camera) {
		Vector2f cVec = new Vector2f(), rVec = new Vector2f(); // chunk, real
		float camX = rVec.x = (float) (camera.camX / World.xMult / Chunk.chunkSize);
		float camZ = rVec.y = MathUtil.superMod(camera.position.z / Chunk.chunkSize, Chunk.ringSize * 2 * HexagonModel.y60);
		
		if (chunks.isEmpty()) {
			MathUtil.realCoordsToChunk(rVec, cVec);
			setNewChunk((int) cVec.x, (int) cVec.y);
		} else {
			for (Iterator<Long> it = chunks.keySet().iterator(); it.hasNext();) {
				long c = it.next();
				int cx = (int) (c >> 32), cz = (int) c, ix, iz;
				cVec.x = cx + 0.5f;
				cVec.y = cz + 0.5f;
				MathUtil.chunkCoordsToReal(cVec, rVec);
				
				float xx, zz, distSq;
				xx = rVec.x - camX;
				zz = Math.min(zz = Math.abs(rVec.y - camZ), Chunk.ringSize * 2 * HexagonModel.y60 - zz); // 'zz =' is important!
				distSq = xx * xx + zz * zz;
				
				if (distSq > viewDistance * viewDistance) {
					Chunk chunk = getChunk(cx, cz);
					chunk.onRemove(true);
					it.remove();
				} else {
					for (int i = 0; i < 4; i++) {
						ix = cx + (2 * i + 1) % 3 - 1;
						iz = cz + (2 * i + 1) / 3 - 1;
						iz = MathUtil.fitChunkZ(iz);
						
						if (!chunks.containsKey((long) ix << 32 | iz)) {
							cVec.x = ix + 0.5f;
							cVec.y = iz + 0.5f;
							MathUtil.chunkCoordsToReal(cVec, rVec);
							xx = rVec.x - camX;
							zz = Math.min(zz = Math.abs(rVec.y - camZ), Chunk.ringSize * 2 * HexagonModel.y60 - zz); // 'zz =' is important!
							distSq = xx * xx + zz * zz;
							
							if (distSq <= viewDistance * viewDistance) {
								setNewChunk(ix, iz);
								return;
							}
						}
					}
				}
			}
		}
	}
	
	private void updateHoverBlock() {
		Main.mousePicker.updateBlockPointedAt(this);
		hoverBlock = Main.mousePicker.getHoverBlock();
		if (hoverBlock != null) {
			if (blockAlteringCountdown == 0) {
				if (Mouse.isButtonDown(0)) {
					removeBlock(hoverBlock.x, hoverBlock.y, hoverBlock.z);
					hoverBlock = null;
					blockAlteringCountdown = 12;
				} else if (Mouse.isButtonDown(1)) {
					int side = Main.mousePicker.getHoverSide();
					if (side != -1) {
						int xx = hoverBlock.x + Block.neighbourOffsets[side][0];
						int yy = hoverBlock.y + Block.neighbourOffsets[side][1];
						int zz = hoverBlock.z + Block.neighbourOffsets[side][2];
						Block block = getBlock(xx, yy, zz);
						if (block == null) {
							setBlock(xx, yy, zz, block = new Block(xx, yy, zz, this));
						}
					}
					blockAlteringCountdown = 12;
				}
			}
		}
		
		if (blockAlteringCountdown > 0) blockAlteringCountdown--;
	}
	
	public void render(Camera camera) {
		Shader.block.enable();
		
		BlockRenderer<Block> blockRenderer;
//		long t1 = 0, tmp1;
		for (int i = 0; i < 8; i++) {
			if ((Main.sides & 1 << i) == 0) continue;
			blockRenderer = i < 2 ? BlockRenderer.block : BlockRenderer.side;
			blockRenderer.begin();
			blockRenderer.submitPerVertexData(i);
//			tmp1 = System.nanoTime();
			
			blockRenderer.submit(blockRenderBuffers[i]);
			
//			Block b;
//			for (Chunk c : chunks.values()) {
//				for (Long l : c.blockRendered.get(i)) {
//					b = c.blocks.get(l);
//					if ((b.sidesToRender & 1 << i) == 0) System.out.println("oops!!");
//					if (b != hoverBlock) blockRenderer.submit(b);
//					num++;
//				}
//			}
//			t1 += System.nanoTime() - tmp1;
//			blockRenderer.end();
//			blockRenderer.render();
		}
//		System.out.println(t1);
		
		if (hoverBlock != null) {
			Shader.block.setUniform1f("glow", 1.5f);
			for (int i = 0; i < 8; i++) {
				if ((Main.sides & 1 << i) == 0 || (hoverBlock.sidesToRender & 1 << i) == 0) continue;
				blockRenderer = i < 2 ? BlockRenderer.block : BlockRenderer.side;
				blockRenderer.begin();
				blockRenderer.submitPerVertexData(i);
				blockRenderer.submit(hoverBlock);
				blockRenderer.end();
				blockRenderer.render();
			}
			Shader.block.setUniform1f("glow", 1f);
		}
	}
	
	public void submitBlock(Block b, ByteBuffer buffer) {
		buffer.putFloat(b.x).putFloat(b.y).putFloat(b.z);
		buffer.putFloat(b.colorTint.x);
		buffer.putFloat(b.colorTint.y);
		buffer.putFloat(b.colorTint.z);
	}
	
	protected void addToRenderList(int list, long block) {
		int N = blockToIndexMaps.get(list).size();
		blockToIndexMaps.get(list).put(block, N);
		indexToBlockLists.get(list).add(block);
		
		if (blockRenderBuffers[list].remaining() < blockRenderStrides[list]) {
			ByteBuffer newBuffer = ByteBuffer.allocate(blockRenderBuffers[list].capacity() * 3 / 2).order(ByteOrder.nativeOrder());
			System.arraycopy(blockRenderBuffers[list].array(), 0, newBuffer.array(), 0, blockRenderBuffers[list].capacity());
			newBuffer.position(blockRenderBuffers[list].position());
			blockRenderBuffers[list] = newBuffer;
		}
		
		submitBlock(blocks.get(block), blockRenderBuffers[list]);
	}
	
	protected void removeFromRenderList(int list, long block) {
		int index = blockToIndexMaps.get(list).remove(block);
		int Nm1 = blockToIndexMaps.get(list).size(); // N - 1
		long last = indexToBlockLists.get(list).remove(Nm1);
		int stride = blockRenderStrides[list];
		if (index != Nm1) {
			indexToBlockLists.get(list).set(index, last);
			blockToIndexMaps.get(list).put(last, index);
			byte[] arr = blockRenderBuffers[list].array();
			System.arraycopy(arr, Nm1 * stride, arr, index * stride, stride);
		}
		blockRenderBuffers[list].position(blockRenderBuffers[list].position() - stride);
	}
	
	public void updateBlock(int x, int y, int z) {
		Block b = blocks.get(MathUtil.encodeCoords(x, y, z));
		byte diff = b.sidesToRender, after;
		b.onUpdate();
		diff ^= after = b.sidesToRender;
		
		if (diff != 0) {
			for (int i = 0; i < 8; i++) {
				if ((diff & 1 << i) != 0) {
					if ((after & 1 << i) != 0) addToRenderList(i, b.coords);
					else removeFromRenderList(i, b.coords);
				}
			}
		}
	}
	
	public Block getBlock(int x, int y, int z) {
		z = MathUtil.superMod(z, Chunk.totalSize);
		
		int cx = Math.floorDiv(x, Chunk.chunkSize), cz = z / Chunk.chunkSize;
		
		Chunk chunk = getChunk(cx, cz);
		return chunk == null ? null : chunk.getBlock(x, y, z);
	}
	
	public void setBlock(int x, int y, int z, Block block) {
		if (z != MathUtil.superMod(z, Chunk.totalSize)) System.out.println("Noo");
		z = MathUtil.superMod(z, Chunk.totalSize);
		
		int cx = Math.floorDiv(x, Chunk.chunkSize), cz = z / Chunk.chunkSize;
		
		Chunk chunk = getChunk(cx, cz);
		if (chunk == null) chunk = setNewChunk(cx, cz);
		chunk.setBlock(x, y, z, block);
	}
	
	public void removeBlock(int x, int y, int z) {
		z = MathUtil.superMod(z, Chunk.totalSize);
		
		int cx = Math.floorDiv(x, Chunk.chunkSize), cz = z / Chunk.chunkSize;
		
		Chunk chunk = getChunk(cx, cz);
		if (chunk != null) chunk.removeBlock(x, y, z);
	}
	
	protected Chunk getChunk(int x, int z) {
		return chunks.get((long) x << 32 | MathUtil.fitChunkZ(z));
	}
	
	private Chunk setNewChunk(int x, int z) {
		z = MathUtil.fitChunkZ(z);
		
		Chunk chunk = new Chunk(x, z, tempBlockGenerator.apply(x, z), this);
		Chunk prev = chunks.put((long) x << 32 | z, chunk);
		if (prev != null) prev.onRemove(false);
		
		return chunk;
	}
	
	public void addBlockToUpdateList(Block b) {
		if (b.willUpdate) return;
		blocksToUpdate.addLast(b);
		b.willUpdate = true;
	}
	
	public void setSaveFolder(File folder) {
		if (!folder.exists()) {
			folder.mkdir();
		}
		saveFolder = folder;
	}
	
	public void saveWorld() {
		if (saveFolder != null) {
			for (Chunk c : chunks.values())
				c.saveChunk();
		}
	}
	
	public void unloadWorld() {
		updateBlocks();
		
		for (Iterator<Chunk> it = chunks.values().iterator(); it.hasNext(); it.remove())
			it.next().onRemove(true);
		
		blocksToUpdate.clear();
		
		blocksToUpdate = null;
		hoverBlock = null;
		chunks = null;
	}
	
}
