package hexagon.world;

import java.io.File;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.channels.FileChannel;
import java.util.function.Consumer;

import org.lwjgl.BufferUtils;
import org.lwjgl.input.Mouse;
import org.lwjgl.util.vector.Vector3f;

import gnu.trove.iterator.TLongIterator;
import gnu.trove.list.array.TLongArrayList;
import gnu.trove.map.hash.TLongIntHashMap;
import gnu.trove.map.hash.TLongObjectHashMap;
import hexagon.Camera;
import hexagon.Main;
import hexagon.graphics.BlockRenderer;
import hexagon.graphics.Shader;
import hexagon.model.HexagonModel;
import hexagon.util.MathUtil;

public class World { // TODO: Move stuff from Chunk and Block to World, and make BlockTypes
	
	public static float viewDistance = 5f; // Change first number only!
	public static float reachDistance = 8f;
	public static final double angleHalfHexagon = Math.PI / Chunk.totalSize;
	public static final double radius = 100;//HexagonModel.y60 / angleHalfHexagon;
	public static final double xMult = radius / (HexagonModel.y60 / angleHalfHexagon);
	public static final float blockToCoordRatio = HexagonModel.y60 * 2;
	
	private TLongObjectHashMap<Chunk> chunks = new TLongObjectHashMap<>();
	private TLongArrayList blocksToUpdate = new TLongArrayList();
	private TLongArrayList blocksUpdating = new TLongArrayList();
	private Block hoverBlock;
	private int blockAlteringCountdown = 0;
	
	// Rendering blocks
	private int[] blockRenderStrides = new int[8];
	private ByteBuffer[] blockRenderBuffers = new ByteBuffer[8];
	private TLongIntHashMap[] blockToIndexMaps = new TLongIntHashMap[8];
	private long[][] indexToBlockLists = new long[8][10000];
	
	// Saving to file
	protected File saveFolder;
	protected ByteBuffer saveBuffer = BufferUtils.createByteBuffer(1024);
	protected long bytes = 0, time = 0;
	protected Consumer<FileChannel> saveBufferFull = (fc) -> {
		try {
			bytes += saveBuffer.position();
			saveBuffer.flip();
			long start = System.nanoTime();
			fc.write(saveBuffer);
			time += System.nanoTime() - start;
			saveBuffer.clear();
			
		} catch (IOException e) {
			e.printStackTrace();
		}
	};
	
	public World() {
		for (int i = 0; i < 8; i++)
			blockRenderStrides[i] = i < 2 ? BlockRenderer.block.instanceSize : BlockRenderer.side.instanceSize;
		for (int i = 0; i < 8; i++)
			blockRenderBuffers[i] = ByteBuffer.allocate(10000 * blockRenderStrides[i]).order(ByteOrder.nativeOrder());
		for (int i = 0; i < 8; i++)
			blockToIndexMaps[i] = new TLongIntHashMap();
	}
	
	public void tick(Camera camera) {
		updateBlocks();
		updateHoverBlock();
	}
	
	private void updateBlocks() {
		TLongArrayList tempList = blocksUpdating;
		blocksUpdating = blocksToUpdate;
		blocksToUpdate = tempList;
		
		long b;
		Block block;
		for (TLongIterator it = blocksUpdating.iterator(); it.hasNext();) {
			b = it.next();
			updateBlock(b);
		}
		
		for (TLongIterator it = blocksUpdating.iterator(); it.hasNext();) {
			b = it.next();
			block = getBlock(b);
			if (block != null) block.willUpdate = false;
		}
		
		blocksUpdating.clear();
	}
	
	// TODO: Rewrite World.updateLoadedChunks() to include y
	public void updateLoadedChunks(Camera camera) {
		long cRem = 0, cAdd = 0;
		float cRemDistSq = -1, cAddDistSq = -1;
		Vector3f cVec = new Vector3f(), rVec = new Vector3f(); // chunk, real
		float camX = rVec.x = camera.position.x / Chunk.chunkSize;
		float camZ = rVec.z = MathUtil.superMod(camera.position.z / Chunk.chunkSize, Chunk.ringSize * 2 * HexagonModel.y60);
		MathUtil.realCoordsToChunk(rVec, cVec);
		
		if (getChunk((int) Math.floor(cVec.x), -1, (int) Math.floor(cVec.z)) == null) {
			setNewChunk((int) Math.floor(cVec.x), -1, (int) Math.floor(cVec.z));
			return;
		} else {
			for (long c : chunks._set) {
				if (!chunks.contains(c)) continue;
				int cx = (int) (c >> 32) << 12 >> 12, cy = (int) (c >> 20) << 20 >> 20, cz = (int) c << 12 >> 12, ix, iz;
				cVec.x = cx + 0.5f;
				cVec.y = cy + 0.5f;
				cVec.z = cz + 0.5f;
				MathUtil.chunkCoordsToReal(cVec, rVec);
				
				float xx, zz, distSq;
				xx = rVec.x - camX;
				zz = Math.min(zz = Math.abs(rVec.z - camZ), Chunk.ringSize * 2 * HexagonModel.y60 - zz); // 'zz =' is important!
				distSq = xx * xx + zz * zz;
				distSq /= blockToCoordRatio * blockToCoordRatio;
				
//				System.out.println(distSq + "\t" + cx + "\t" + cz + "\t" + (c >> 32));
				if (distSq > viewDistance * viewDistance) {
					if (cRemDistSq < 0 || distSq > cRemDistSq) {
						cRem = encodeCoords(cx, cy, cz);
						cRemDistSq = distSq;
					}
				} else {
					for (int i = 0; i < 4; i++) {
						ix = cx + (2 * i + 1) % 3 - 1;
						iz = cz + (2 * i + 1) / 3 - 1;
						iz = MathUtil.fitChunkZ(iz);
						
						if (!chunks.containsKey(encodeCoords(ix, -1, iz))) {
							cVec.x = ix + 0.5f;
							cVec.z = iz + 0.5f;
							MathUtil.chunkCoordsToReal(cVec, rVec);
							xx = rVec.x - camX;
							zz = Math.min(zz = Math.abs(rVec.z - camZ), Chunk.ringSize * 2 * HexagonModel.y60 - zz); // 'zz =' is important!
							distSq = xx * xx + zz * zz;
							distSq /= blockToCoordRatio * blockToCoordRatio;
							
							if (distSq <= viewDistance * viewDistance) {
								if (cAddDistSq < 0 || distSq < cAddDistSq) {
									cAdd = encodeCoords(ix, -1, iz);
									cAddDistSq = distSq;
								}
							}
						}
					}
				}
			}
		}
		
		if (cRemDistSq >= 0) {
			Chunk c = getChunk(cRem);
			if (c != null) c.onRemove(true);
			chunks.remove(cRem);
		}
		if (cAddDistSq >= 0) setNewChunk((int) (cAdd >> 32) << 12 >> 12, (int) (cAdd >> 20) << 20 >> 20, (int) cAdd << 12 >> 12);
	}
	
	private void updateHoverBlock() {
		Main.mousePicker.updateBlockPointedAt(this);
		hoverBlock = Main.mousePicker.getHoverBlock();
		if (hoverBlock != null) {
			if (blockAlteringCountdown == 0) {
				if (Mouse.isButtonDown(0)) {
					removeBlock(hoverBlock.x(), hoverBlock.y(), hoverBlock.z());
					hoverBlock = null;
					blockAlteringCountdown = 12;
				} else if (Mouse.isButtonDown(1)) {
					int side = Main.mousePicker.getHoverSide();
					if (side != -1) {
						int xx = hoverBlock.x() + Block.neighbourOffsets[side][0];
						int yy = hoverBlock.y() + Block.neighbourOffsets[side][1];
						int zz = hoverBlock.z() + Block.neighbourOffsets[side][2];
						zz = MathUtil.superMod(zz, Chunk.totalSize);
						Block block = getBlock(xx, yy, zz);
						if (block == null) {
							Chunk ch = getChunkForSure(xx >> 4, yy >> 4, zz >> 4);
							ch.setBlock(block = new Block(xx, yy, zz, ch));
						}
					}
					blockAlteringCountdown = 12;
				}
			}
		}
		
		if (blockAlteringCountdown > 0) blockAlteringCountdown--;
	}
	
	public void render(Camera camera) {
		Shader.block.enable();
		
		BlockRenderer<Block> blockRenderer;
//		long t1 = 0, tmp1;
		for (int i = 0; i < 8; i++) {
			if ((Main.sides & 1 << i) == 0) continue;
			blockRenderer = i < 2 ? BlockRenderer.block : BlockRenderer.side;
			blockRenderer.begin();
			blockRenderer.submitPerVertexData(i);
//			tmp1 = System.nanoTime();
			
			blockRenderer.submit(blockRenderBuffers[i]);
			
//			Block b;
//			for (Chunk c : chunks.values()) {
//				for (Long l : c.blockRendered.get(i)) {
//					b = c.blocks.get(l);
//					if ((b.sidesToRender & 1 << i) == 0) System.out.println("oops!!");
//					if (b != hoverBlock) blockRenderer.submit(b);
//					num++;
//				}
//			}
//			t1 += System.nanoTime() - tmp1;
//			blockRenderer.end();
//			blockRenderer.render();
		}
//		System.out.println(t1);
		
		if (hoverBlock != null) {
			Shader.block.setUniform1f("glow", 1.5f);
			for (int i = 0; i < 8; i++) {
				if ((Main.sides & 1 << i) == 0 || (hoverBlock.sidesToRender & 1 << i) == 0) continue;
				blockRenderer = i < 2 ? BlockRenderer.block : BlockRenderer.side;
				blockRenderer.begin();
				blockRenderer.submitPerVertexData(i);
				blockRenderer.submit(hoverBlock);
				blockRenderer.end();
				blockRenderer.render();
			}
			Shader.block.setUniform1f("glow", 1f);
		}
	}
	
	public void submitBlock(Block b, ByteBuffer buffer) {
		if (b == null) {
			buffer.position(buffer.position() + 24);
			return;
		}
		buffer.putFloat(b.x()).putFloat(b.y()).putFloat(b.z());
		buffer.putFloat(b.colorTintR);
		buffer.putFloat(b.colorTintG);
		buffer.putFloat(b.colorTintB);
	}
	
	protected void addToRenderList(int list, long block) {
		int N = blockToIndexMaps[list].size();
		blockToIndexMaps[list].put(block, N);
		if (N == indexToBlockLists[list].length) {
			long[] newArr = new long[N * 3 / 2];
			System.arraycopy(indexToBlockLists[list], 0, newArr, 0, N);
			indexToBlockLists[list] = newArr;
			
			ByteBuffer newBuffer = ByteBuffer.allocate(indexToBlockLists[list].length * blockRenderStrides[list]).order(ByteOrder.nativeOrder());
			System.arraycopy(blockRenderBuffers[list].array(), 0, newBuffer.array(), 0, blockRenderBuffers[list].capacity());
			newBuffer.position(blockRenderBuffers[list].position());
			blockRenderBuffers[list] = newBuffer;
		}
		indexToBlockLists[list][N] = block;
		
		if (indexToBlockLists[list].length * blockRenderStrides[list] != blockRenderBuffers[list].capacity()) System.out.println(indexToBlockLists[list].length * blockRenderStrides[list] + "\t" + blockRenderBuffers[list].capacity() + "\tA");
		
		submitBlock(getBlock(block), blockRenderBuffers[list]);
	}
	
	protected void removeFromRenderList(int list, long block) {
		if (indexToBlockLists[list].length * blockRenderStrides[list] != blockRenderBuffers[list].capacity()) System.out.println(indexToBlockLists[list].length * blockRenderStrides[list] + "\t" + blockRenderBuffers[list].capacity());
		int index = blockToIndexMaps[list].remove(block);
		int Nm1 = blockToIndexMaps[list].size(); // N - 1
		long last = indexToBlockLists[list][Nm1];
		int stride = blockRenderStrides[list];
		if (index != Nm1) {
			indexToBlockLists[list][index] = last;
			blockToIndexMaps[list].put(last, index);
			byte[] arr = blockRenderBuffers[list].array();
			try {
				System.arraycopy(arr, Nm1 * stride, arr, index * stride, stride);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		blockRenderBuffers[list].position(blockRenderBuffers[list].position() - stride);
	}
	
	public void updateBlock(int x, int y, int z) {
		updateBlock(World.encodeBlockCoords(x >> 4, y >> 4, z >> 4, Chunk.encodeCoords(x, y, z)));
	}
	
	public void updateBlock(long coords) {
		Block b = getBlock(coords);
		if (b == null) return;
		byte diff = b.sidesToRender, after;
		b.onUpdate();
		diff ^= after = b.sidesToRender;
		
		if (diff != 0) {
			for (int i = 0; i < 8; i++) {
				if ((diff & 1 << i) != 0) {
					if ((after & 1 << i) != 0) addToRenderList(i, coords);
					else removeFromRenderList(i, coords);
				}
			}
		}
	}
	
	public Block getBlock(long block) {
		Chunk c = getChunk(block >>> 12);
		return c == null ? null : c.getBlock((short) (block & 0xfff));
	}
	
	public Block getBlock(int x, int y, int z) {
		z = MathUtil.superMod(z, Chunk.totalSize);
		
		Chunk chunk = getChunk(x >> 4, y >> 4, z >> 4);
		return chunk == null ? null : chunk.getBlock(x, y, z);
	}
	
	public void setBlock(Block block) {
		block.chunk.setBlock(block);
	}
	
	public void removeBlock(int x, int y, int z) {
		z = MathUtil.superMod(z, Chunk.totalSize);
		
		int cx = Math.floorDiv(x, Chunk.chunkSize), cy = Math.floorDiv(y, Chunk.chunkSize), cz = z / Chunk.chunkSize;
		
		Chunk chunk = getChunk(cx, cy, cz);
		if (chunk != null) chunk.removeBlock(x, y, z);
	}
	
	protected Chunk getChunk(long chunk) {
		return chunks.get(chunk);
	}
	
	protected Chunk getChunk(int x, int y, int z) {
		return chunks.get(encodeCoords(x, y, z));
	}
	
	public Chunk getChunkForSure(int x, int y, int z) {
		z = MathUtil.fitChunkZ(z);
		
		Chunk chunk = chunks.get(encodeCoords(x, y, z));
		if (chunk == null) {
			chunk = new Chunk(x, y, z, this);
			Chunk prev = chunks.put(encodeCoords(x, y, z), chunk);
			if (prev != null) prev.onRemove(false);
		}
		
		return chunk;
	}
	
	private Chunk setNewChunk(int x, int y, int z) {
		z = MathUtil.fitChunkZ(z);
		
		Chunk chunk = new Chunk(x, y, z, this);
		Chunk prev = chunks.put(encodeCoords(x, y, z), chunk);
		if (prev != null) prev.onRemove(false);
		
		return chunk;
	}
	
	public static long encodeCoords(int x, int y, int z) { // bytes in long: xxxxxyyy zzzzz
		return (x & 0xfffffL) << 32 | (y & 0xfffL) << 20 | z & 0xfffffL;
	}
	
	public static long encodeBlockCoords(int x, int y, int z, short blockCoords) { // bytes in long: xxxxxyyy zzzzzbbb where b is blockcoords xyz
		return encodeCoords(x, y, z) << 12 | blockCoords;
	}
	
	public static long encodeBlockCoords(long chunk, short blockCoords) { // bytes in long: xxxxxyyy zzzzzbbb where b is blockcoords xyz
		return chunk << 12 | blockCoords;
	}
	
	public static long encodeBlockCoords(int x, int y, int z) { // bytes in long: xxxxxyyy zzzzzbbb where b is blockcoords xyz
		return encodeBlockCoords(x >> 4, y >> 4, MathUtil.superMod(z, Chunk.totalSize) >> 4, Chunk.encodeCoords(x, y, z));
	}
	
	public void addBlockToUpdateList(long coords, Block b) {
		if (b.willUpdate) return;
		blocksToUpdate.add(coords);
		b.willUpdate = true;
	}
	
	public void setSaveFolder(File folder) {
		if (!folder.exists()) {
			folder.mkdir();
		}
		saveFolder = folder;
	}
	
	public void saveWorld() {
		if (saveFolder != null) {
			chunks.forEachValue(c -> {
				c.saveChunk();
				return true;
			});
		}
	}
	
	public void unloadWorld() {
		updateBlocks();
		
		chunks.forEachValue(c -> {
			c.onRemove(true);
			return true;
		});
		
		blocksToUpdate.clear();
		
		blocksToUpdate = null;
		hoverBlock = null;
		chunks = null;
	}
	
}
